@@ -1,4 +1,6 @@
 from __future__ import print_function, absolute_import
+
+import contextlib
 import functools
 
 from . import instructions, types, values
@@ -70,6 +72,10 @@ class IRBuilder(object):
     def function(self):
         return self.block.parent
 
+    @property
+    def module(self):
+        return self.block.parent.module
+
     def position_before(self, instr):
         self._block = instr.parent
         self._anchor = self._block.instructions.index(instr)
@@ -86,6 +92,96 @@ class IRBuilder(object):
         self._block = block
         self._anchor = len(block.instructions)
 
+    def append_basic_block(self, name=''):
+        return self.function.append_basic_block(name)
+
+    @contextlib.contextmanager
+    def goto_block(self, block):
+        """
+        A context manager which temporarily positions the builder at the end
+        of basic block *bb* (but before any terminator).
+        """
+        old_block = self.basic_block
+        term = block.terminator
+        if term is not None:
+            self.position_before(term)
+        else:
+            self.position_at_end(block)
+        try:
+            yield
+        finally:
+            self.position_at_end(old_block)
+
+    @contextlib.contextmanager
+    def goto_entry_block(self):
+        """
+        A context manager which temporarily positions the builder at the
+        end of the function's entry block.
+        """
+        with self.goto_block(self.function.entry_basic_block):
+            yield
+
+    @contextlib.contextmanager
+    def _branch_helper(self, bbenter, bbexit):
+        self.position_at_end(bbenter)
+        yield bbexit
+        if bbenter.terminator is None:
+            self.branch(bbexit)
+
+    @contextlib.contextmanager
+    def if_then(self, pred, likely=None):
+        """
+        A context manager which sets up a conditional basic block based
+        on the given predicate (a i1 value).  If the conditional block
+        is not explicitly terminated, a branch will be added to the next
+        block.
+        If *likely* is given, its boolean value indicates whether the
+        predicate is likely to be true or not, and metadata is issued
+        for LLVM's optimizers to account for that.
+        """
+        bb = self.basic_block
+        bbif = self.append_basic_block(name=bb.name + '.if')
+        bbend = self.append_basic_block(name=bb.name + '.endif')
+        br = self.cbranch(pred, bbif, bbend)
+        if likely is not None:
+            br.set_weights([99, 1] if likely else [1, 99])
+
+        with self._branch_helper(bbif, bbend):
+            yield bbend
+
+        self.position_at_end(bbend)
+
+    @contextlib.contextmanager
+    def if_else(self, pred, likely=None):
+        """
+        A context manager which sets up two conditional basic blocks based
+        on the given predicate (a i1 value).
+        A tuple of context managers is yield'ed.  Each context manager
+        acts as a if_then() block.
+        *likely* has the same meaning as in if_then().
+
+        Typical use::
+            with builder.if_else(pred) as (then, otherwise):
+                with then:
+                    # emit instructions for when the predicate is true
+                with otherwise:
+                    # emit instructions for when the predicate is false
+        """
+        bb = self.basic_block
+        bbif = self.append_basic_block(name=bb.name + '.if')
+        bbelse = self.append_basic_block(name=bb.name + '.else')
+        bbend = self.append_basic_block(name=bb.name + '.endif')
+        br = self.cbranch(pred, bbif, bbelse)
+        if likely is not None:
+            br.set_weights([99, 1] if likely else [1, 99])
+
+        then = self._branch_helper(bbif, bbend)
+        otherwise = self._branch_helper(bbelse, bbend)
+
+        yield then, otherwise
+
+        self.position_at_end(bbend)
+
     def constant(self, typ, val):
         return values.Constant(typ, val)
 
@@ -281,6 +377,10 @@ class IRBuilder(object):
     def inttoptr(self, value, typ, name=''):
         pass
 
+    @_castop('addrspacecast')
+    def addrspacecast(self, value, typ, name=''):
+        pass
+
     #
     # Memory APIs
     #
@@ -307,11 +407,11 @@ class IRBuilder(object):
         self._insert(ld)
         return ld
 
-    def store(self, val, ptr):
+    def store(self, value, ptr):
         if not isinstance(ptr.type, types.PointerType):
             raise TypeError("cannot store to value of type %s (%r): not a pointer"
                             % (ptr.type, str(ptr)))
-        st = instructions.StoreInstr(self.block, val, ptr)
+        st = instructions.StoreInstr(self.block, value, ptr)
         self._insert(st)
         return st
 
@@ -320,24 +420,25 @@ class IRBuilder(object):
     # Terminators APIs
     #
 
-    def switch(self, val, elseblk):
-        swt = instructions.SwitchInstr(self.block, 'switch', val, elseblk)
+    def switch(self, value, default):
+        swt = instructions.SwitchInstr(self.block, 'switch', value, default)
         self._set_terminator(swt)
         return swt
 
     def branch(self, target):
         """Jump to target
         """
-        term = instructions.Terminator(self.block, "br", [target])
-        self._set_terminator(term)
-        return term
+        br = instructions.Branch(self.block, "br", [target])
+        self._set_terminator(br)
+        return br
 
     def cbranch(self, cond, truebr, falsebr):
         """Branch conditionally
         """
-        term = instructions.Terminator(self.block, "br", [cond, truebr, falsebr])
-        self._set_terminator(term)
-        return term
+        br = instructions.ConditionalBranch(self.block, "br",
+                                            [cond, truebr, falsebr])
+        self._set_terminator(br)
+        return br
 
     def ret_void(self):
         return self._set_terminator(
@@ -349,8 +450,9 @@ class IRBuilder(object):
 
     # Call APIs
 
-    def call(self, fn, args, name='', cconv=None):
-        inst = instructions.CallInstr(self.block, fn, args, name=name, cconv=cconv)
+    def call(self, fn, args, name='', cconv=None, tail=False):
+        inst = instructions.CallInstr(self.block, fn, args, name=name,
+                                      cconv=cconv, tail=tail)
         self._insert(inst)
         return inst
 
@@ -371,10 +473,10 @@ class IRBuilder(object):
         self._insert(instr)
         return instr
 
-    def insert_value(self, agg, elem, idx, name=''):
+    def insert_value(self, agg, value, idx, name=''):
         if not isinstance(idx, (tuple, list)):
             idx = [idx]
-        instr = instructions.InsertValue(self.block, agg, elem, idx, name=name)
+        instr = instructions.InsertValue(self.block, agg, value, idx, name=name)
         self._insert(instr)
         return instr
 
